#define THROWING_STATIC_ASSERT // Make static asserts throw, to make testing easier
#include "bounded_int.hpp"

#include "catch.hpp"

static constexpr int intmax = std::numeric_limits<int>::max();
static constexpr int intmin = std::numeric_limits<int>::min();

using namespace provided::literals;

template <intmax_t L, intmax_t R>
int get_value(bounded_int<L, R> const& bi) {
    return static_cast<int>(bi);
}

template <intmax_t L, intmax_t R>
int get_min(bounded_int<L, R> const&) {
    return L;
}

template <intmax_t L, intmax_t R>
int get_max(bounded_int<L, R> const&) {
    return R;
}

TEST_CASE("Literals and constructors") {
    SECTION("int constructor") {
        bounded_int<1, 3> bi(1);
        REQUIRE(get_min(bi) == 1);
        REQUIRE(get_max(bi) == 3);
        REQUIRE(get_value(bi) == 1);
        REQUIRE_THROWS_AS((bounded_int<1, 4>(0)), std::domain_error);
    }
    SECTION("Literal") {
        auto bi1 = 1_bi;
        REQUIRE(get_min(bi1) == 1);
        REQUIRE(get_max(bi1) == 1);
        REQUIRE(get_value(bi1) == 1);
        auto bi2 = 123456_bi;
        REQUIRE(get_min(bi2) == 123456);
        REQUIRE(get_max(bi2) == 123456);
        REQUIRE(get_value(bi2) == 123456);
    }
    SECTION("Widening copy operations are OK") {
        bounded_int<1, 3> bi(1);
        bounded_int<0, 5> wider(bi);
        REQUIRE(get_value(wider) == 1);
        wider = bi;
        REQUIRE(get_value(wider) == 1);
    }
    SECTION("Narrowing copy operations are invalid") {
        bounded_int<1, 3> bi(1);
        REQUIRE_THROWS_AS((bounded_int<1, 1>{bi}), provided::static_exception);
        bounded_int<1, 1> narrow(1);
        REQUIRE_THROWS_AS(narrow = bi, provided::static_exception);
    }
}

TEST_CASE("Addition") {
    SECTION("Basic inference") {
        auto bi1 = bounded_int<1, 2> {1} +  bounded_int<1, 3> {3};
        REQUIRE(get_min(bi1) == 2);
        REQUIRE(get_max(bi1) == 5);
        REQUIRE(get_value(bi1) == 4);
        auto bi2 = bounded_int<-7, -2>{-4} + bounded_int<2, 4>{3};
        REQUIRE(get_min(bi2) == -5);
        REQUIRE(get_max(bi2) == 2);
        REQUIRE(get_value(bi2) == -1);
        auto bi3 = bounded_int<-1000, 1000>{0} + bounded_int<-1000, 1000>{0};
        REQUIRE(get_min(bi3) == -2000);
        REQUIRE(get_max(bi3) == 2000);
        REQUIRE(get_value(bi3) == 0);
        auto bi4 = bounded_int<0, intmax - 1>{ 1 } +bounded_int<-2, 1>{0};
        REQUIRE(get_min(bi4) == -2);
        REQUIRE(get_max(bi4) == intmax);
        REQUIRE(get_value(bi4) == 1);
    }
    SECTION("Invalid operations") {
        REQUIRE_THROWS_AS((bounded_int<0, 1'500'000'000>{1} +bounded_int<-1, 1'500'000'000>{0}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<0, intmax>{1} +bounded_int<-1, 1>{0}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<intmin, 0>{-1} +bounded_int<-1, 1>{0}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<-1'500'000'000, 0>{-3} +bounded_int<-1'500'000'000, 123> {12}), provided::static_exception);
    }
}

TEST_CASE("Subtraction") {
    SECTION("Basic inference") {
        auto bi1 = bounded_int<1, 12>{4} - bounded_int<0, 2>{1};
        REQUIRE(get_min(bi1) == -1);
        REQUIRE(get_max(bi1) == 12);
        REQUIRE(get_value(bi1) == 3);
        auto bi2 = bounded_int<-2, 5>{ 0 } -bounded_int<-10, 2>{0};
        REQUIRE(get_min(bi2) == -4);
        REQUIRE(get_max(bi2) == 15);
        REQUIRE(get_value(bi2) == 0);
        auto bi3 = bounded_int<-7, 21>{ -5 } -bounded_int <22, 73>{25};
        REQUIRE(get_min(bi3) == -80);
        REQUIRE(get_max(bi3) == -1);
        REQUIRE(get_value(bi3) == -30);
        auto bi4 = bounded_int<1, 7>{ 3 } -bounded_int<-22, -2>{-5};
        REQUIRE(get_min(bi4) == 3);
        REQUIRE(get_max(bi4) == 29);
        REQUIRE(get_value(bi4) == 8);
    }
    SECTION("Invalid operations") {
        REQUIRE_THROWS_AS((bounded_int<0, intmax>{0} - bounded_int<intmin, 0>{0}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<0, 2'000'000'000>{3} -bounded_int<-1'000'000'000, 123>{23}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<intmin, 0>{0} - bounded_int<0, intmax>{0}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<-2'000'000'000, -123>{-144} -bounded_int<333, 1'000'000'000>{444}), provided::static_exception);
    }
}

TEST_CASE("Negation (unary -)") {
    SECTION("Basic inference") {
        auto bi = bounded_int<-13, -1>{-3};
        auto bi2 = -bi;
        REQUIRE(get_min(bi2) == 1);
        REQUIRE(get_max(bi2) == 13);
        REQUIRE(get_value(bi2) == 3);
        auto bi3 = bounded_int<22, 24>{ 23 };
        auto bi4 = -bi3;
        REQUIRE(get_min(bi4) == -24);
        REQUIRE(get_max(bi4) == -22);
        REQUIRE(get_value(bi4) == -23);
    }
    SECTION("Invalid operations") {
        bounded_int<intmin, 0> bi1{ -2 };
        REQUIRE_THROWS_AS(-bi1, provided::static_exception);
    }
}

TEST_CASE("Unary +") {
    SECTION("Basic inference") {
        auto bi = bounded_int<-123, -12>{ -15 };
        auto bi2 = +bi;
        REQUIRE(get_min(bi2) == -123);
        REQUIRE(get_max(bi2) == -12);
        REQUIRE(get_value(bi2) == -15);
    }
}

TEST_CASE("Multiplication") {
    SECTION("Basic inference") {
        auto bi = bounded_int<-100, 10>{ -1 } *bounded_int<-100, 100>{23};
        REQUIRE(get_min(bi) == -10'000);
        REQUIRE(get_max(bi) == 10'000);
        REQUIRE(get_value(bi) == -23);
        auto bi2 = bounded_int<-123, 123>{ 122 } *bounded_int<-123, 124>{0};
        REQUIRE(get_min(bi2) == -15252);
        REQUIRE(get_max(bi2) == 15252);
        REQUIRE(get_value(bi2) == 0);
        auto bi3 = bounded_int<-1, 2>{ 1 } *bounded_int<12, 123>{111};
        REQUIRE(get_min(bi3) == -123);
        REQUIRE(get_max(bi3) == 246);
        REQUIRE(get_value(bi3) == 111);
        auto bi4 = bounded_int<-22, 23>{ -2 } *bounded_int<-1, 1>{0};
        REQUIRE(get_min(bi4) == -23);
        REQUIRE(get_max(bi4) == 23);
        REQUIRE(get_value(bi4) == 0);
        auto bi5 = bounded_int<-12, -1>{ -3 } *bounded_int<-123, -2>{-5};
        REQUIRE(get_min(bi5) == 2);
        REQUIRE(get_max(bi5) == 1476);
        REQUIRE(get_value(bi5) == 15);
        auto bi6 = bounded_int<7, 664>{ 10 } *bounded_int<4, 774>{10};
        REQUIRE(get_min(bi6) == 28);
        REQUIRE(get_max(bi6) == 664 * 774);
        REQUIRE(get_value(bi6) == 100);
        REQUIRE_NOTHROW(bounded_int<-23, 46340> { 2} *bounded_int<-2, 46340>{2});
    }
    SECTION("Invalid operations") {
        REQUIRE_THROWS_AS((bounded_int<-1'000'000, -1>{ -3 } * bounded_int<10, 10'000>{ 32 }), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<-200'000, 25>{2} *bounded_int<-200'000, 33>{4}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<1, 1'000'000>{ 23 } *bounded_int<1'000, 10'000>{ 1001 }), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<-74, 46360>{3} *bounded_int<-23, 46432>{5}), provided::static_exception);
    }
}

TEST_CASE("Division") {
    SECTION("Basic inference") {
        auto bi = bounded_int<123, 1234>{140} / bounded_int<12, 24>{14};
        REQUIRE(get_min(bi) == 5);
        REQUIRE(get_max(bi) == 102);
        REQUIRE(get_value(bi) == 10);
        auto bi2 = bounded_int<-1234, -123>{ -140 } / bounded_int<12, 24>{14};
        REQUIRE(get_min(bi2) == -102);
        REQUIRE(get_max(bi2) == -5);
        REQUIRE(get_value(bi2) == -10);
    }
    SECTION("Invalid division") {
        REQUIRE_THROWS_AS((bounded_int<123, 1234>{144} / bounded_int<-12, 12>{2}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<123, 1234>{1111} / bounded_int<-1, 12>{11}), provided::static_exception);
    }
}

TEST_CASE("Comparison operators") {
    SECTION("Proper comparisons") {
        REQUIRE(bounded_int<12, 123>{ 23 } < bounded_int<12, 123>{ 24 });
        REQUIRE_FALSE(bounded_int<12, 14>{14} < bounded_int<14, 15>{14});

        REQUIRE(bounded_int<12, 123>{25} > bounded_int<12, 123>{23});
        REQUIRE_FALSE(bounded_int<21, 22>{21} > bounded_int<21, 27>{24});

        REQUIRE(bounded_int<12, 123>{33} <= bounded_int<12, 123>{33});
        REQUIRE_FALSE(bounded_int<12, 14>{13} <= bounded_int<10, 12>{10});

        REQUIRE(bounded_int<12, 123>{100} >= bounded_int<12, 123>{15});
        REQUIRE_FALSE(bounded_int<12, 123>{100} >= bounded_int<12, 123>{101});

        REQUIRE(bounded_int<12, 35>{30} == bounded_int<27, 33>{30});
        REQUIRE_FALSE(bounded_int<12, 33>{30} == bounded_int<33, 35>{33});

        REQUIRE(bounded_int<19, 43>{21} != bounded_int<0, 19>{19});
        REQUIRE_FALSE(bounded_int<21, 27>{23} != bounded_int<22, 24>{23});
    }
    SECTION("Statically determinable comparisons") {
        REQUIRE_THROWS_AS((bounded_int<12, 14>{13} < bounded_int<100, 120>{110}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<14, 15>{14} < bounded_int<12, 14>{13}), provided::static_exception);

        REQUIRE_THROWS_AS((bounded_int<21, 22>{21} > bounded_int<22, 27>{24}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<123, 124>{123} > bounded_int<-2, 2>{0}), provided::static_exception);

        REQUIRE_THROWS_AS((bounded_int<12, 123>{33} <= bounded_int<124, 164>{157}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<12, 123>{33} <= bounded_int<0, 11>{7}), provided::static_exception);

        REQUIRE_THROWS_AS((bounded_int<12, 123>{100} >= bounded_int<0, 11>{10}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<12, 123>{100} >= bounded_int<1111, 2222>{1234}), provided::static_exception);

        REQUIRE_THROWS_AS((bounded_int<12, 30>{30} == bounded_int<31, 33>{31}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<30, 30>{30} != bounded_int<31, 31>{31}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<30, 30>{30} == bounded_int<30, 30>{30}), provided::static_exception);
        REQUIRE_THROWS_AS((bounded_int<30, 30>{30} != bounded_int<30, 30>{30}), provided::static_exception);
    }
}
